@@ -20,8 +20,7 @@
  */
 
 /*
- * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
- * Use is subject to license terms.
+ * Copyright (c) 2002, 2010, Oracle and/or its affiliates. All rights reserved.
  */
 
 #include <stdio.h>
@@ -61,8 +60,8 @@ static struct uuid_to_ptag {
 	{ EFI_RESERVED },
 	{ EFI_SYSTEM },
 	{ EFI_LEGACY_MBR },
-	{ EFI_RESV3 },
-	{ EFI_RESV4 },
+	{ EFI_SYMC_PUB },
+	{ EFI_SYMC_CDS },
 	{ EFI_MSFT_RESV },
 	{ EFI_DELL_BASIC },
 	{ EFI_DELL_RAID },
@@ -88,7 +87,8 @@ struct dk_map2  default_vtoc_map[NDKMAP] = {
 
 #if defined(_SUNOS_VTOC_16)
 
-#if defined(i386) || defined(__amd64)
+#if defined(i386) || defined(__amd64) || defined(__arm) || \
+    defined(__powerpc) || defined(__sparc)
 	{	V_BOOT,		V_UNMNT	},		/* i - 8 */
 	{	V_ALTSCTR,	0	},		/* j - 9 */
 
@@ -133,8 +133,8 @@ read_disk_info(int fd, diskaddr_t *capacity, uint_t *lbsize)
 	int sector_size;
 	unsigned long long capacity_size;
 
-        if (ioctl(fd, BLKSSZGET, &sector_size) < 0)
-                return (-1);
+	if (ioctl(fd, BLKSSZGET, &sector_size) < 0)
+		return (-1);
 
 	if (ioctl(fd, BLKGETSIZE64, &capacity_size) < 0)
 		return (-1);
@@ -153,7 +153,7 @@ efi_get_info(int fd, struct dk_cinfo *dki_info)
 	char *dev_path;
 	int rval = 0;
 
-	memset(dki_info, 0, sizeof(*dki_info));
+	memset(dki_info, 0, sizeof (*dki_info));
 
 	path = calloc(PATH_MAX, 1);
 	if (path == NULL)
@@ -183,38 +183,48 @@ efi_get_info(int fd, struct dk_cinfo *dki_info)
 		strcpy(dki_info->dki_cname, "sd");
 		dki_info->dki_ctype = DKC_SCSI_CCS;
 		rval = sscanf(dev_path, "/dev/%[a-zA-Z]%hu",
-			      dki_info->dki_dname,
-			      &dki_info->dki_partition);
+		    dki_info->dki_dname,
+		    &dki_info->dki_partition);
 	} else if ((strncmp(dev_path, "/dev/hd", 7) == 0)) {
 		strcpy(dki_info->dki_cname, "hd");
 		dki_info->dki_ctype = DKC_DIRECT;
 		rval = sscanf(dev_path, "/dev/%[a-zA-Z]%hu",
-			      dki_info->dki_dname,
-			      &dki_info->dki_partition);
+		    dki_info->dki_dname,
+		    &dki_info->dki_partition);
 	} else if ((strncmp(dev_path, "/dev/md", 7) == 0)) {
 		strcpy(dki_info->dki_cname, "pseudo");
 		dki_info->dki_ctype = DKC_MD;
-		rval = sscanf(dev_path, "/dev/%[a-zA-Z0-9]p%hu",
-			      dki_info->dki_dname,
-			      &dki_info->dki_partition);
+		strcpy(dki_info->dki_dname, "md");
+		rval = sscanf(dev_path, "/dev/md%[0-9]p%hu",
+		    dki_info->dki_dname + 2,
+		    &dki_info->dki_partition);
+	} else if ((strncmp(dev_path, "/dev/vd", 7) == 0)) {
+		strcpy(dki_info->dki_cname, "vd");
+		dki_info->dki_ctype = DKC_MD;
+		rval = sscanf(dev_path, "/dev/%[a-zA-Z]%hu",
+		    dki_info->dki_dname,
+		    &dki_info->dki_partition);
 	} else if ((strncmp(dev_path, "/dev/dm-", 8) == 0)) {
 		strcpy(dki_info->dki_cname, "pseudo");
 		dki_info->dki_ctype = DKC_VBD;
-		rval = sscanf(dev_path, "/dev/%[a-zA-Z0-9-]p%hu",
-			      dki_info->dki_dname,
-			      &dki_info->dki_partition);
+		strcpy(dki_info->dki_dname, "dm-");
+		rval = sscanf(dev_path, "/dev/dm-%[0-9]p%hu",
+		    dki_info->dki_dname + 3,
+		    &dki_info->dki_partition);
 	} else if ((strncmp(dev_path, "/dev/ram", 8) == 0)) {
 		strcpy(dki_info->dki_cname, "pseudo");
 		dki_info->dki_ctype = DKC_PCMCIA_MEM;
-		rval = sscanf(dev_path, "/dev/%[a-zA-Z0-9]p%hu",
-			      dki_info->dki_dname,
-			      &dki_info->dki_partition);
+		strcpy(dki_info->dki_dname, "ram");
+		rval = sscanf(dev_path, "/dev/ram%[0-9]p%hu",
+		    dki_info->dki_dname + 3,
+		    &dki_info->dki_partition);
 	} else if ((strncmp(dev_path, "/dev/loop", 9) == 0)) {
 		strcpy(dki_info->dki_cname, "pseudo");
 		dki_info->dki_ctype = DKC_VBD;
-		rval = sscanf(dev_path, "/dev/%[a-zA-Z0-9]p%hu",
-			      dki_info->dki_dname,
-			      &dki_info->dki_partition);
+		strcpy(dki_info->dki_dname, "loop");
+		rval = sscanf(dev_path, "/dev/loop%[0-9]p%hu",
+		    dki_info->dki_dname + 4,
+		    &dki_info->dki_partition);
 	} else {
 		strcpy(dki_info->dki_dname, "unknown");
 		strcpy(dki_info->dki_cname, "unknown");
@@ -270,19 +280,12 @@ efi_alloc_and_init(int fd, uint32_t nparts, struct dk_gpt **vtoc)
 	struct uuid	uuid;
 	struct dk_cinfo	dki_info;
 
-	if (read_disk_info(fd, &capacity, &lbsize) != 0) {
-		if (efi_debug)
-			(void) fprintf(stderr,
-			    "couldn't read disk information\n");
+	if (read_disk_info(fd, &capacity, &lbsize) != 0)
 		return (-1);
-	}
+
 #if defined(__linux__)
-	if (efi_get_info(fd, &dki_info) != 0) {
-		if (efi_debug)
-			(void) fprintf(stderr,
-			    "couldn't read disk information\n");
+	if (efi_get_info(fd, &dki_info) != 0)
 		return (-1);
-	}
 
 	if (dki_info.dki_partition != 0)
 		return (-1);
@@ -397,10 +400,10 @@ efi_ioctl(int fd, int cmd, dk_efi_t *dk_ioc)
 	 */
 	if (read_disk_info(fd, &capacity, &lbsize) == -1) {
 		if (efi_debug)
-			fprintf(stderr,"unable to read disk info: %d",errno);
+			fprintf(stderr, "unable to read disk info: %d", errno);
 
 		errno = EIO;
-		return -1;
+		return (-1);
 	}
 
 	switch (cmd) {
@@ -408,7 +411,7 @@ efi_ioctl(int fd, int cmd, dk_efi_t *dk_ioc)
 		if (lbsize == 0) {
 			if (efi_debug)
 				(void) fprintf(stderr, "DKIOCGETEFI assuming "
-					       "LBA %d bytes\n", DEV_BSIZE);
+				    "LBA %d bytes\n", DEV_BSIZE);
 
 			lbsize = DEV_BSIZE;
 		}
@@ -417,24 +420,24 @@ efi_ioctl(int fd, int cmd, dk_efi_t *dk_ioc)
 		if (error == -1) {
 			if (efi_debug)
 				(void) fprintf(stderr, "DKIOCGETEFI lseek "
-				               "error: %d\n", errno);
-			return error;
+				    "error: %d\n", errno);
+			return (error);
 		}
 
 		error = read(fd, data, dk_ioc->dki_length);
 		if (error == -1) {
 			if (efi_debug)
 				(void) fprintf(stderr, "DKIOCGETEFI read "
-				               "error: %d\n", errno);
-			return error;
+				    "error: %d\n", errno);
+			return (error);
 		}
 
 		if (error != dk_ioc->dki_length) {
 			if (efi_debug)
 				(void) fprintf(stderr, "DKIOCGETEFI short "
-					       "read of %d bytes\n", error);
+				    "read of %d bytes\n", error);
 			errno = EIO;
-			return -1;
+			return (-1);
 		}
 		error = 0;
 		break;
@@ -443,43 +446,43 @@ efi_ioctl(int fd, int cmd, dk_efi_t *dk_ioc)
 		if (lbsize == 0) {
 			if (efi_debug)
 				(void) fprintf(stderr, "DKIOCSETEFI unknown "
-					       "LBA size\n");
+				    "LBA size\n");
 			errno = EIO;
-			return -1;
+			return (-1);
 		}
 
 		error = lseek(fd, dk_ioc->dki_lba * lbsize, SEEK_SET);
 		if (error == -1) {
 			if (efi_debug)
 				(void) fprintf(stderr, "DKIOCSETEFI lseek "
-				               "error: %d\n", errno);
-			return error;
+				    "error: %d\n", errno);
+			return (error);
 		}
 
 		error = write(fd, data, dk_ioc->dki_length);
 		if (error == -1) {
 			if (efi_debug)
 				(void) fprintf(stderr, "DKIOCSETEFI write "
-				               "error: %d\n", errno);
-			return error;
+				    "error: %d\n", errno);
+			return (error);
 		}
 
 		if (error != dk_ioc->dki_length) {
 			if (efi_debug)
 				(void) fprintf(stderr, "DKIOCSETEFI short "
-					       "write of %d bytes\n", error);
+				    "write of %d bytes\n", error);
 			errno = EIO;
-			return -1;
+			return (-1);
 		}
 
 		/* Sync the new EFI table to disk */
 		error = fsync(fd);
 		if (error == -1)
-			return error;
+			return (error);
 
 		/* Ensure any local disk cache is also flushed */
 		if (ioctl(fd, BLKFLSBUF, 0) == -1)
-			return error;
+			return (error);
 
 		error = 0;
 		break;
@@ -489,7 +492,7 @@ efi_ioctl(int fd, int cmd, dk_efi_t *dk_ioc)
 			(void) fprintf(stderr, "unsupported ioctl()\n");
 
 		errno = EIO;
-		return -1;
+		return (-1);
 	}
 #else
 	dk_ioc->dki_data_64 = (uint64_t)(uintptr_t)data;
@@ -499,10 +502,10 @@ efi_ioctl(int fd, int cmd, dk_efi_t *dk_ioc)
 	return (error);
 }
 
-#if defined(__linux__)
-static int
+int
 efi_rescan(int fd)
 {
+#if defined(__linux__)
 	int retry = 5;
 	int error;
 
@@ -510,14 +513,14 @@ efi_rescan(int fd)
 	while ((error = ioctl(fd, BLKRRPART)) != 0) {
 		if (--retry == 0) {
 			(void) fprintf(stderr, "the kernel failed to rescan "
-				       "the partition table: %d\n", errno);
+			    "the partition table: %d\n", errno);
 			return (-1);
 		}
 	}
+#endif
 
 	return (0);
 }
-#endif
 
 static int
 check_label(int fd, dk_efi_t *dk_ioc)
@@ -549,16 +552,22 @@ check_label(int fd, dk_efi_t *dk_ioc)
 	 */
 	crc = efi->efi_gpt_HeaderCRC32;
 	efi->efi_gpt_HeaderCRC32 = 0;
+	len_t headerSize = (len_t)LE_32(efi->efi_gpt_HeaderSize);
 
-	if (((len_t)LE_32(efi->efi_gpt_HeaderSize) > dk_ioc->dki_length) ||
-	    crc != LE_32(efi_crc32((unsigned char *)efi,
-	    LE_32(efi->efi_gpt_HeaderSize)))) {
+	if (headerSize < EFI_MIN_LABEL_SIZE || headerSize > EFI_LABEL_SIZE) {
+		if (efi_debug)
+			(void) fprintf(stderr,
+				"Invalid EFI HeaderSize %llu.  Assuming %d.\n",
+				headerSize, EFI_MIN_LABEL_SIZE);
+	}
+
+	if ((headerSize > dk_ioc->dki_length) ||
+	    crc != LE_32(efi_crc32((unsigned char *)efi, headerSize))) {
 		if (efi_debug)
 			(void) fprintf(stderr,
 			    "Bad EFI CRC: 0x%x != 0x%x\n",
-			    crc,
-			    LE_32(efi_crc32((unsigned char *)efi,
-			    sizeof (struct efi_gpt))));
+			    crc, LE_32(efi_crc32((unsigned char *)efi,
+			    headerSize)));
 		return (VT_EINVAL);
 	}
 
@@ -587,7 +596,7 @@ efi_read(int fd, struct dk_gpt *vtoc)
 	 * get the partition number for this file descriptor.
 	 */
 	if ((rval = efi_get_info(fd, &dki_info)) != 0)
-		return rval;
+		return (rval);
 
 	if ((strncmp(dki_info.dki_cname, "pseudo", 7) == 0) &&
 	    (strncmp(dki_info.dki_dname, "md", 3) == 0)) {
@@ -605,8 +614,8 @@ efi_read(int fd, struct dk_gpt *vtoc)
 	if (read_disk_info(fd, &capacity, &lbsize) == -1) {
 		if (efi_debug) {
 			(void) fprintf(stderr,
-				       "unable to read disk info: %d",
-				       errno);
+			    "unable to read disk info: %d",
+			    errno);
 		}
 		return (VT_EINVAL);
 	}
@@ -639,7 +648,7 @@ efi_read(int fd, struct dk_gpt *vtoc)
 	}
 
 	if (posix_memalign((void **)&dk_ioc.dki_data,
-		           disk_info.dki_lbsize, label_len))
+	    disk_info.dki_lbsize, label_len))
 		return (VT_ERROR);
 
 	memset(dk_ioc.dki_data, 0, label_len);
@@ -1024,24 +1033,15 @@ efi_use_whole_disk(int fd)
 	struct dk_gpt		*efi_label;
 	int			rval;
 	int			i;
-	uint_t			phy_last_slice = 0;
-	diskaddr_t		pl_start = 0;
-	diskaddr_t		pl_size;
+	uint_t			resv_index = 0, data_index = 0;
+	diskaddr_t		resv_start = 0, data_start = 0;
+	diskaddr_t		difference;
 
 	rval = efi_alloc_and_read(fd, &efi_label);
 	if (rval < 0) {
 		return (rval);
 	}
 
-	/* find the last physically non-zero partition */
-	for (i = 0; i < efi_label->efi_nparts - 2; i ++) {
-		if (pl_start < efi_label->efi_parts[i].p_start) {
-			pl_start = efi_label->efi_parts[i].p_start;
-			phy_last_slice = i;
-		}
-	}
-	pl_size = efi_label->efi_parts[phy_last_slice].p_size;
-
 	/*
 	 * If alter_lba is 1, we are using the backup label.
 	 * Since we can locate the backup label by disk capacity,
@@ -1057,16 +1057,28 @@ efi_use_whole_disk(int fd)
 		return (VT_ENOSPC);
 	}
 
+	difference = efi_label->efi_last_lba - efi_label->efi_altern_lba;
+
 	/*
-	 * If there is space between the last physically non-zero partition
-	 * and the reserved partition, just add the unallocated space to this
-	 * area. Otherwise, the unallocated space is added to the last
-	 * physically non-zero partition.
+	 * Find the last physically non-zero partition.
+	 * This is the reserved partition.
 	 */
-	if (pl_start + pl_size - 1 == efi_label->efi_last_u_lba -
-	    EFI_MIN_RESV_SIZE) {
-		efi_label->efi_parts[phy_last_slice].p_size +=
-		    efi_label->efi_last_lba - efi_label->efi_altern_lba;
+	for (i = 0; i < efi_label->efi_nparts; i ++) {
+		if (resv_start < efi_label->efi_parts[i].p_start) {
+			resv_start = efi_label->efi_parts[i].p_start;
+			resv_index = i;
+		}
+	}
+
+	/*
+	 * Find the last physically non-zero partition before that.
+	 * This is the data partition.
+	 */
+	for (i = 0; i < resv_index; i ++) {
+		if (data_start < efi_label->efi_parts[i].p_start) {
+			data_start = efi_label->efi_parts[i].p_start;
+			data_index = i;
+		}
 	}
 
 	/*
@@ -1074,10 +1086,9 @@ efi_use_whole_disk(int fd)
 	 * here except fabricated devids (which get generated via
 	 * efi_write()). So there is no need to copy data.
 	 */
-	efi_label->efi_parts[efi_label->efi_nparts - 1].p_start +=
-	    efi_label->efi_last_lba - efi_label->efi_altern_lba;
-	efi_label->efi_last_u_lba += efi_label->efi_last_lba
-	    - efi_label->efi_altern_lba;
+	efi_label->efi_parts[data_index].p_size += difference;
+	efi_label->efi_parts[resv_index].p_start += difference;
+	efi_label->efi_last_u_lba += difference;
 
 	rval = efi_write(fd, efi_label);
 	if (rval < 0) {
@@ -1112,7 +1123,7 @@ efi_write(int fd, struct dk_gpt *vtoc)
 	diskaddr_t		lba_backup_gpt_hdr;
 
 	if ((rval = efi_get_info(fd, &dki_info)) != 0)
-		return rval;
+		return (rval);
 
 	/* check if we are dealing wih a metadevice */
 	if ((strncmp(dki_info.dki_cname, "pseudo", 7) == 0) &&
@@ -1151,7 +1162,7 @@ efi_write(int fd, struct dk_gpt *vtoc)
 	 */
 	lba_backup_gpt_hdr = vtoc->efi_last_u_lba + 1 + nblocks;
 	if (posix_memalign((void **)&dk_ioc.dki_data,
-		           vtoc->efi_lbasize, dk_ioc.dki_length))
+	    vtoc->efi_lbasize, dk_ioc.dki_length))
 		return (VT_ERROR);
 
 	memset(dk_ioc.dki_data, 0, dk_ioc.dki_length);
@@ -1160,7 +1171,7 @@ efi_write(int fd, struct dk_gpt *vtoc)
 	/* stuff user's input into EFI struct */
 	efi->efi_gpt_Signature = LE_64(EFI_SIGNATURE);
 	efi->efi_gpt_Revision = LE_32(vtoc->efi_version); /* 0x02000100 */
-	efi->efi_gpt_HeaderSize = LE_32(sizeof (struct efi_gpt));
+	efi->efi_gpt_HeaderSize = LE_32(sizeof (struct efi_gpt) - LEN_EFI_PAD);
 	efi->efi_gpt_Reserved1 = 0;
 	efi->efi_gpt_MyLBA = LE_64(1ULL);
 	efi->efi_gpt_AlternateLBA = LE_64(lba_backup_gpt_hdr);
@@ -1229,7 +1240,8 @@ efi_write(int fd, struct dk_gpt *vtoc)
 	    LE_32(efi_crc32((unsigned char *)efi_parts,
 	    vtoc->efi_nparts * (int)sizeof (struct efi_gpe)));
 	efi->efi_gpt_HeaderCRC32 =
-	    LE_32(efi_crc32((unsigned char *)efi, sizeof (struct efi_gpt)));
+	    LE_32(efi_crc32((unsigned char *)efi,
+	    LE_32(efi->efi_gpt_HeaderSize)));
 
 	if (efi_ioctl(fd, DKIOCSETEFI, &dk_ioc) == -1) {
 		free(dk_ioc.dki_data);
@@ -1282,7 +1294,7 @@ efi_write(int fd, struct dk_gpt *vtoc)
 	efi->efi_gpt_HeaderCRC32 = 0;
 	efi->efi_gpt_HeaderCRC32 =
 	    LE_32(efi_crc32((unsigned char *)dk_ioc.dki_data,
-	    sizeof (struct efi_gpt)));
+	    LE_32(efi->efi_gpt_HeaderSize)));
 
 	if (efi_ioctl(fd, DKIOCSETEFI, &dk_ioc) == -1) {
 		if (efi_debug) {
@@ -1297,12 +1309,6 @@ efi_write(int fd, struct dk_gpt *vtoc)
 	(void) write_pmbr(fd, vtoc);
 	free(dk_ioc.dki_data);
 
-#if defined(__linux__)
-	rval = efi_rescan(fd);
-	if (rval)
-		return (VT_ERROR);
-#endif
-
 	return (0);
 }
 
