@@ -27,31 +27,21 @@
 #include_next <unistd.h>
 
 #ifndef _LIBSPL_UNISTD_H
-#define _LIBSPL_UNISTD_H
+#define	_LIBSPL_UNISTD_H
 
 #if !defined(HAVE_IOCTL_IN_UNISTD_H)
-# if defined(HAVE_IOCTL_IN_SYS_IOCTL_H)
-#  include <sys/ioctl.h>
-# elif defined(HAVE_IOCTL_IN_STROPTS_H)
-#  include <stropts.h>
-# else
-#  error "System call ioctl() unavailable"
-# endif
-#endif
+#if defined(HAVE_IOCTL_IN_SYS_IOCTL_H)
+#include <sys/ioctl.h>
+#elif defined(HAVE_IOCTL_IN_STROPTS_H)
+#include <stropts.h>
+#else /* HAVE_IOCTL_IN_STROPTS_H */
+#error "System call ioctl() unavailable"
+#endif /* HAVE_IOCTL_IN_SYS_IOCTL_H */
+#endif /* !HAVE_IOCTL_IN_UNISTD_H */
 
 #if !defined(HAVE_ISSETUGID)
-# include <sys/types.h>
-# define issetugid() (geteuid() == 0 || getegid() == 0)
-#endif
-
-#if !defined(__sun__) && !defined(__sun)
-/* It seems Solaris only returns positive host ids */
-static inline long fake_gethostid(void)
-{
-	long id = gethostid();
-	return id >= 0 ? id : -id;
-}
-#define gethostid() fake_gethostid()
+#include <sys/types.h>
+#define	issetugid() (geteuid() == 0 || getegid() == 0)
 #endif
 
 #endif /* _LIBSPL_UNISTD_H */
