@@ -76,14 +76,22 @@
 #include <sys/time.h>
 #include <sys/zfs_ioctl.h>
 
-int zevent_len_max = 0;
-int zevent_cols = 80;
-int zevent_console = 0;
+int zfs_zevent_len_max = 0;
+int zfs_zevent_cols = 80;
+int zfs_zevent_console = 0;
 
 static int zevent_len_cur = 0;
 static int zevent_waiters = 0;
 static int zevent_flags = 0;
 
+/*
+ * The EID (Event IDentifier) is used to uniquely tag a zevent when it is
+ * posted.  The posted EIDs are monotonically increasing but not persistent.
+ * They will be reset to the initial value (1) each time the kernel module is
+ * loaded.
+ */
+static uint64_t zevent_eid = 0;
+
 static kmutex_t zevent_lock;
 static list_t zevent_list;
 static kcondvar_t zevent_cv;
@@ -276,8 +284,8 @@ fm_nvprintr(nvlist_t *nvl, int d, int c, int cols)
 			c = fm_printf(d + 1, c, cols, "[ ");
 			(void) nvpair_value_int8_array(nvp, &val, &nelem);
 			for (i = 0; i < nelem; i++)
-			        c = fm_printf(d + 1, c, cols, "0x%llx ",
-					      (u_longlong_t)val[i]);
+				c = fm_printf(d + 1, c, cols, "0x%llx ",
+				    (u_longlong_t)val[i]);
 
 			c = fm_printf(d + 1, c, cols, "]");
 			break;
@@ -290,8 +298,8 @@ fm_nvprintr(nvlist_t *nvl, int d, int c, int cols)
 			c = fm_printf(d + 1, c, cols, "[ ");
 			(void) nvpair_value_uint8_array(nvp, &val, &nelem);
 			for (i = 0; i < nelem; i++)
-			        c = fm_printf(d + 1, c, cols, "0x%llx ",
-					      (u_longlong_t)val[i]);
+				c = fm_printf(d + 1, c, cols, "0x%llx ",
+				    (u_longlong_t)val[i]);
 
 			c = fm_printf(d + 1, c, cols, "]");
 			break;
@@ -304,8 +312,8 @@ fm_nvprintr(nvlist_t *nvl, int d, int c, int cols)
 			c = fm_printf(d + 1, c, cols, "[ ");
 			(void) nvpair_value_int16_array(nvp, &val, &nelem);
 			for (i = 0; i < nelem; i++)
-			        c = fm_printf(d + 1, c, cols, "0x%llx ",
-					      (u_longlong_t)val[i]);
+				c = fm_printf(d + 1, c, cols, "0x%llx ",
+				    (u_longlong_t)val[i]);
 
 			c = fm_printf(d + 1, c, cols, "]");
 			break;
@@ -318,8 +326,8 @@ fm_nvprintr(nvlist_t *nvl, int d, int c, int cols)
 			c = fm_printf(d + 1, c, cols, "[ ");
 			(void) nvpair_value_uint16_array(nvp, &val, &nelem);
 			for (i = 0; i < nelem; i++)
-			        c = fm_printf(d + 1, c, cols, "0x%llx ",
-					      (u_longlong_t)val[i]);
+				c = fm_printf(d + 1, c, cols, "0x%llx ",
+				    (u_longlong_t)val[i]);
 
 			c = fm_printf(d + 1, c, cols, "]");
 			break;
@@ -332,8 +340,8 @@ fm_nvprintr(nvlist_t *nvl, int d, int c, int cols)
 			c = fm_printf(d + 1, c, cols, "[ ");
 			(void) nvpair_value_int32_array(nvp, &val, &nelem);
 			for (i = 0; i < nelem; i++)
-			        c = fm_printf(d + 1, c, cols, "0x%llx ",
-					      (u_longlong_t)val[i]);
+			c = fm_printf(d + 1, c, cols, "0x%llx ",
+			    (u_longlong_t)val[i]);
 
 			c = fm_printf(d + 1, c, cols, "]");
 			break;
@@ -346,8 +354,8 @@ fm_nvprintr(nvlist_t *nvl, int d, int c, int cols)
 			c = fm_printf(d + 1, c, cols, "[ ");
 			(void) nvpair_value_uint32_array(nvp, &val, &nelem);
 			for (i = 0; i < nelem; i++)
-			        c = fm_printf(d + 1, c, cols, "0x%llx ",
-					      (u_longlong_t)val[i]);
+				c = fm_printf(d + 1, c, cols, "0x%llx ",
+				    (u_longlong_t)val[i]);
 
 			c = fm_printf(d + 1, c, cols, "]");
 			break;
@@ -360,8 +368,8 @@ fm_nvprintr(nvlist_t *nvl, int d, int c, int cols)
 			c = fm_printf(d + 1, c, cols, "[ ");
 			(void) nvpair_value_int64_array(nvp, &val, &nelem);
 			for (i = 0; i < nelem; i++)
-			        c = fm_printf(d + 1, c, cols, "0x%llx ",
-					      (u_longlong_t)val[i]);
+				c = fm_printf(d + 1, c, cols, "0x%llx ",
+				    (u_longlong_t)val[i]);
 
 			c = fm_printf(d + 1, c, cols, "]");
 			break;
@@ -374,8 +382,8 @@ fm_nvprintr(nvlist_t *nvl, int d, int c, int cols)
 			c = fm_printf(d + 1, c, cols, "[ ");
 			(void) nvpair_value_uint64_array(nvp, &val, &nelem);
 			for (i = 0; i < nelem; i++)
-			        c = fm_printf(d + 1, c, cols, "0x%llx ",
-					      (u_longlong_t)val[i]);
+				c = fm_printf(d + 1, c, cols, "0x%llx ",
+				    (u_longlong_t)val[i]);
 
 			c = fm_printf(d + 1, c, cols, "]");
 			break;
@@ -405,9 +413,9 @@ fm_nvprint(nvlist_t *nvl)
 	console_printf("\n");
 
 	if (nvlist_lookup_string(nvl, FM_CLASS, &class) == 0)
-		c = fm_printf(0, c, zevent_cols, "%s", class);
+		c = fm_printf(0, c, zfs_zevent_cols, "%s", class);
 
-	if (fm_nvprintr(nvl, 0, c, zevent_cols) != 0)
+	if (fm_nvprintr(nvl, 0, c, zfs_zevent_cols) != 0)
 		console_printf("\n");
 
 	console_printf("\n");
@@ -418,15 +426,15 @@ zfs_zevent_alloc(void)
 {
 	zevent_t *ev;
 
-	ev = kmem_zalloc(sizeof(zevent_t), KM_SLEEP);
+	ev = kmem_zalloc(sizeof (zevent_t), KM_PUSHPAGE);
 	if (ev == NULL)
-		return NULL;
+		return (NULL);
 
-	list_create(&ev->ev_ze_list, sizeof(zfs_zevent_t),
+	list_create(&ev->ev_ze_list, sizeof (zfs_zevent_t),
 		    offsetof(zfs_zevent_t, ze_node));
 	list_link_init(&ev->ev_node);
 
-	return ev;
+	return (ev);
 }
 
 static void
@@ -436,7 +444,7 @@ zfs_zevent_free(zevent_t *ev)
 	ev->ev_cb(ev->ev_nvl, ev->ev_detector);
 
 	list_destroy(&ev->ev_ze_list);
-	kmem_free(ev, sizeof(zevent_t));
+	kmem_free(ev, sizeof (zevent_t));
 }
 
 static void
@@ -481,14 +489,13 @@ zfs_zevent_drain_all(int *count)
 static void
 zfs_zevent_insert(zevent_t *ev)
 {
-	mutex_enter(&zevent_lock);
+	ASSERT(MUTEX_HELD(&zevent_lock));
 	list_insert_head(&zevent_list, ev);
-	if (zevent_len_cur >= zevent_len_max)
+
+	if (zevent_len_cur >= zfs_zevent_len_max)
 		zfs_zevent_drain(list_tail(&zevent_list));
 	else
 		zevent_len_cur++;
-
-	mutex_exit(&zevent_lock);
 }
 
 /*
@@ -499,6 +506,7 @@ zfs_zevent_post(nvlist_t *nvl, nvlist_t *detector, zevent_cb_t *cb)
 {
 	int64_t tv_array[2];
 	timestruc_t tv;
+	uint64_t eid;
 	size_t nvl_size = 0;
 	zevent_t *ev;
 
@@ -510,13 +518,19 @@ zfs_zevent_post(nvlist_t *nvl, nvlist_t *detector, zevent_cb_t *cb)
 		return;
 	}
 
+	eid = atomic_inc_64_nv(&zevent_eid);
+	if (nvlist_add_uint64(nvl, FM_EREPORT_EID, eid)) {
+		atomic_add_64(&erpt_kstat_data.erpt_set_failed.value.ui64, 1);
+		return;
+	}
+
 	(void) nvlist_size(nvl, &nvl_size, NV_ENCODE_NATIVE);
 	if (nvl_size > ERPT_DATA_SZ || nvl_size == 0) {
 		atomic_add_64(&erpt_kstat_data.erpt_dropped.value.ui64, 1);
 		return;
 	}
 
-	if (zevent_console)
+	if (zfs_zevent_console)
 		fm_nvprint(nvl);
 
 	ev = zfs_zevent_alloc();
@@ -525,11 +539,15 @@ zfs_zevent_post(nvlist_t *nvl, nvlist_t *detector, zevent_cb_t *cb)
 		return;
 	}
 
-        ev->ev_nvl = nvl;
+	ev->ev_nvl = nvl;
 	ev->ev_detector = detector;
 	ev->ev_cb = cb;
+	ev->ev_eid = eid;
+
+	mutex_enter(&zevent_lock);
 	zfs_zevent_insert(ev);
 	cv_broadcast(&zevent_cv);
+	mutex_exit(&zevent_lock);
 }
 
 static int
@@ -548,12 +566,12 @@ zfs_zevent_fd_hold(int fd, minor_t *minorp, zfs_zevent_t **ze)
 	file_t *fp;
 	int error;
 
-        fp = getf(fd);
-        if (fp == NULL)
-                return (EBADF);
+	fp = getf(fd);
+	if (fp == NULL)
+		return (EBADF);
 
-        *minorp = zfsdev_getminor(fp->f_file);
-        error = zfs_zevent_minor_to_state(*minorp, ze);
+	*minorp = zfsdev_getminor(fp->f_file);
+	error = zfs_zevent_minor_to_state(*minorp, ze);
 
 	if (error)
 		zfs_zevent_fd_rele(fd);
@@ -568,13 +586,18 @@ zfs_zevent_fd_rele(int fd)
 }
 
 /*
- * Get the next zevent in the stream and place a copy in 'event'.
+ * Get the next zevent in the stream and place a copy in 'event'.  This
+ * may fail with ENOMEM if the encoded nvlist size exceeds the passed
+ * 'event_size'.  In this case the stream pointer is not advanced and
+ * and 'event_size' is set to the minimum required buffer size.
  */
 int
-zfs_zevent_next(zfs_zevent_t *ze, nvlist_t **event, uint64_t *dropped)
+zfs_zevent_next(zfs_zevent_t *ze, nvlist_t **event, uint64_t *event_size,
+    uint64_t *dropped)
 {
 	zevent_t *ev;
-	int error;
+	size_t size;
+	int error = 0;
 
 	mutex_enter(&zevent_lock);
 	if (ze->ze_zevent == NULL) {
@@ -585,17 +608,27 @@ zfs_zevent_next(zfs_zevent_t *ze, nvlist_t **event, uint64_t *dropped)
 			goto out;
 		}
 	} else {
-		/* Existing stream continue with the next element and remove
-		 * ourselves from the wait queue for the previous element */
+		/*
+		 * Existing stream continue with the next element and remove
+		 * ourselves from the wait queue for the previous element
+		 */
 		ev = list_prev(&zevent_list, ze->ze_zevent);
 		if (ev == NULL) {
 			error = ENOENT;
 			goto out;
 		}
-
-		list_remove(&ze->ze_zevent->ev_ze_list, ze);
 	}
 
+	VERIFY(nvlist_size(ev->ev_nvl, &size, NV_ENCODE_NATIVE) == 0);
+	if (size > *event_size) {
+		*event_size = size;
+		error = ENOMEM;
+		goto out;
+	}
+
+	if (ze->ze_zevent)
+		list_remove(&ze->ze_zevent->ev_ze_list, ze);
+
 	ze->ze_zevent = ev;
 	list_insert_head(&ev->ev_ze_list, ze);
 	nvlist_dup(ev->ev_nvl, event, KM_SLEEP);
@@ -604,7 +637,7 @@ zfs_zevent_next(zfs_zevent_t *ze, nvlist_t **event, uint64_t *dropped)
 out:
 	mutex_exit(&zevent_lock);
 
-	return error;
+	return (error);
 }
 
 int
@@ -628,7 +661,68 @@ zfs_zevent_wait(zfs_zevent_t *ze)
 out:
 	mutex_exit(&zevent_lock);
 
-	return error;
+	return (error);
+}
+
+/*
+ * The caller may seek to a specific EID by passing that EID.  If the EID
+ * is still available in the posted list of events the cursor is positioned
+ * there.  Otherwise ENOENT is returned and the cursor is not moved.
+ *
+ * There are two reserved EIDs which may be passed and will never fail.
+ * ZEVENT_SEEK_START positions the cursor at the start of the list, and
+ * ZEVENT_SEEK_END positions the cursor at the end of the list.
+ */
+int
+zfs_zevent_seek(zfs_zevent_t *ze, uint64_t eid)
+{
+	zevent_t *ev;
+	int error = 0;
+
+	mutex_enter(&zevent_lock);
+
+	if (eid == ZEVENT_SEEK_START) {
+		if (ze->ze_zevent)
+			list_remove(&ze->ze_zevent->ev_ze_list, ze);
+
+		ze->ze_zevent = NULL;
+		goto out;
+	}
+
+	if (eid == ZEVENT_SEEK_END) {
+		if (ze->ze_zevent)
+			list_remove(&ze->ze_zevent->ev_ze_list, ze);
+
+		ev = list_head(&zevent_list);
+		if (ev) {
+			ze->ze_zevent = ev;
+			list_insert_head(&ev->ev_ze_list, ze);
+		} else {
+			ze->ze_zevent = NULL;
+		}
+
+		goto out;
+	}
+
+	for (ev = list_tail(&zevent_list); ev != NULL;
+	    ev = list_prev(&zevent_list, ev)) {
+		if (ev->ev_eid == eid) {
+			if (ze->ze_zevent)
+				list_remove(&ze->ze_zevent->ev_ze_list, ze);
+
+			ze->ze_zevent = ev;
+			list_insert_head(&ev->ev_ze_list, ze);
+			break;
+		}
+	}
+
+	if (ev == NULL)
+		error = ENOENT;
+
+out:
+	mutex_exit(&zevent_lock);
+
+	return (error);
 }
 
 void
@@ -659,7 +753,7 @@ zfs_zevent_destroy(zfs_zevent_t *ze)
 static void *
 i_fm_alloc(nv_alloc_t *nva, size_t size)
 {
-	return (kmem_zalloc(size, KM_SLEEP));
+	return (kmem_zalloc(size, KM_PUSHPAGE));
 }
 
 /* ARGSUSED */
@@ -727,7 +821,7 @@ fm_nvlist_create(nv_alloc_t *nva)
 	nv_alloc_t *nvhdl;
 
 	if (nva == NULL) {
-		nvhdl = kmem_zalloc(sizeof (nv_alloc_t), KM_SLEEP);
+		nvhdl = kmem_zalloc(sizeof (nv_alloc_t), KM_PUSHPAGE);
 
 		if (nv_alloc_init(nvhdl, &fm_mem_alloc_ops, NULL, 0) != 0) {
 			kmem_free(nvhdl, sizeof (nv_alloc_t));
@@ -1400,7 +1494,13 @@ fm_ena_generate_cpu(uint64_t timestamp, processorid_t cpuid, uchar_t format)
 uint64_t
 fm_ena_generate(uint64_t timestamp, uchar_t format)
 {
-	return (fm_ena_generate_cpu(timestamp, getcpuid(), format));
+	uint64_t ena;
+
+	kpreempt_disable();
+	ena = fm_ena_generate_cpu(timestamp, getcpuid(), format);
+	kpreempt_enable();
+
+	return (ena);
 }
 
 uint64_t
@@ -1475,8 +1575,8 @@ fm_init(void)
 	zevent_len_cur = 0;
 	zevent_flags = 0;
 
-	if (zevent_len_max == 0)
-		zevent_len_max = ERPT_MAX_ERRS * MAX(max_ncpus, 4);
+	if (zfs_zevent_len_max == 0)
+		zfs_zevent_len_max = ERPT_MAX_ERRS * MAX(max_ncpus, 4);
 
 	/* Initialize zevent allocation and generation kstats */
 	fm_ksp = kstat_create("zfs", 0, "fm", "misc", KSTAT_TYPE_NAMED,
@@ -1491,7 +1591,8 @@ fm_init(void)
 	}
 
 	mutex_init(&zevent_lock, NULL, MUTEX_DEFAULT, NULL);
-	list_create(&zevent_list, sizeof(zevent_t), offsetof(zevent_t, ev_node));
+	list_create(&zevent_list, sizeof (zevent_t),
+	    offsetof(zevent_t, ev_node));
 	cv_init(&zevent_cv, NULL, CV_DEFAULT, NULL);
 }
 
@@ -1501,9 +1602,10 @@ fm_fini(void)
 	int count;
 
 	zfs_zevent_drain_all(&count);
-	cv_broadcast(&zevent_cv);
 
 	mutex_enter(&zevent_lock);
+	cv_broadcast(&zevent_cv);
+
 	zevent_flags |= ZEVENT_SHUTDOWN;
 	while (zevent_waiters > 0) {
 		mutex_exit(&zevent_lock);
@@ -1522,13 +1624,13 @@ fm_fini(void)
 	}
 }
 
-module_param(zevent_len_max, int, 0644);
-MODULE_PARM_DESC(zevent_len_max, "Maximum event queue length");
+module_param(zfs_zevent_len_max, int, 0644);
+MODULE_PARM_DESC(zfs_zevent_len_max, "Max event queue length");
 
-module_param(zevent_cols, int, 0644);
-MODULE_PARM_DESC(zevent_cols, "Maximum event column width");
+module_param(zfs_zevent_cols, int, 0644);
+MODULE_PARM_DESC(zfs_zevent_cols, "Max event column width");
 
-module_param(zevent_console, int, 0644);
-MODULE_PARM_DESC(zevent_console, "Log events to the console");
+module_param(zfs_zevent_console, int, 0644);
+MODULE_PARM_DESC(zfs_zevent_console, "Log events to the console");
 
 #endif /* _KERNEL */
